from time import sleep

from .config import DISABLE_LABEL, ENABLE_LABEL
from .notifiers import TemplateMessage
from ..update.container import Container
from ..update.service import Service


class Scanner(object):
    def __init__(self, docker_client):
        self.docker = docker_client
        self.logger = self.docker.logger
        self.config = self.docker.config
        self.client = self.docker.client
        self.socket = self.docker.socket
        self.notification_manager = self.docker.notification_manager

    def _scan_containers(self):
        """Return filtered container objects list"""
        monitored_containers = []

        for container in self.client.containers.list(filters={'status': 'running'}, ignore_removed=True):
            enable_label = container.labels.get(ENABLE_LABEL, False)
            disable_label = container.labels.get(DISABLE_LABEL, False)
            swarm = container.labels.get('com.docker.stack.namespace', False)
            if (not self.config.label or enable_label) and not disable_label and not swarm:
                monitored_containers.append(Container(self.docker, container))
            else:
                self.logger.debug("skip monitoring for %s", container.name)

        self.logger.info("total containers monitored: %s", len(monitored_containers))

        return monitored_containers

    def _scan_services(self):
        """Return filtered service objects list"""
        monitored_services = []

        for service in self.client.services.list():
            enable_label = service.attrs['Spec']['Labels'].get(ENABLE_LABEL, False)
            disable_label = service.attrs['Spec']['Labels'].get(DISABLE_LABEL, False)
            if (not self.config.label or enable_label) and not disable_label:
                monitored_services.append(Service(self.docker, service))

        self.logger.info("total services monitored: %s", len(monitored_services))

        return monitored_services

    def scan_monitored(self):
        """Return all object update if there are a new version"""
        monitored = []

        if not self.config.disable_containers_check:
            monitored.extend(self._scan_containers())

        if not self.config.disable_services_check:
            monitored.extend(self._scan_services())

        return monitored

    def check_swarm_mode(self):
        try:
            if not self.client.swarm or not self.client.swarm.attrs:
                self.logger.info("Your aren't running in swarm mode, skip services check.")
                self.config.disable_services_check = True
        except Exception as e:
            error_str = str(e).lower()
            if "this node is not a swarm manager" in error_str:
                if "worker nodes" in error_str:
                    raise EnvironmentError(
                        "Your running on a swarm worker, it isn't working. You must add placement constraints. "
                        "See docs at https://github.com/dockupdater/dockupdater for help."
                    )
                else:
                    self.logger.info("Your aren't running in swarm mode, skip services check.")
                    self.config.disable_services_check = True
            else:
                raise e

        if self.config.disable_containers_check and self.config.disable_services_check:
            raise AttributeError("Error you can't disable all monitoring (containers/services).")

    def update(self):
        monitored = self.scan_monitored()

        if not monitored:
            self.logger.info('No containers/services are running or monitored on %s', self.socket)
            return

        for container_or_service in monitored:
            self.logger.debug("checking object %s", container_or_service.name)

            if container_or_service.has_new_version():
                self.logger.info('%s will be updated', container_or_service.name)
                container_or_service.update()
                self.logger.debug('%s is updated', container_or_service.name)
                self.notification_manager.send(
                    TemplateMessage(container_or_service),
                    container_or_service.config.notifiers
                )
                if container_or_service.config.wait:
                    self.logger.info('Waiting %s before next update', container_or_service.config.wait)
                    sleep(container_or_service.config.wait)
            else:
                self.logger.debug("no new version for %s", container_or_service.name)

    def self_update(self):
        if not self.config.disable_containers_check:
            # Removing old dockupdater
            self.logger.debug('Looking for old dockupdater on %s', self.socket)
            for container in self.client.containers.list(all=True, ignore_removed=True):
                if container.name.endswith("_old_dockupdater"):
                    self.logger.debug('Stopping and deleting %s', container.name)
                    container.stop()
                    container.remove()
            # Fix dockupdater if they need exposed port
            for container in self.client.containers.list(all=True, ignore_removed=True):
                if container.labels.get("dockupdater.updater_port"):
                    self.logger.info('Recreate dockupdater container with exposed ports')
                    Container(self.docker, container)
                    container.update()



from abc import ABC, abstractmethod

from docker.errors import APIError


class AbstractObject(ABC):
    def __init__(self, docker_client, object):
        self.docker = docker_client
        self.logger = self.docker.logger
        self.config = self.docker.config
        self.client = self.docker.client
        self.socket = self.docker.socket

        self.object = object

    @property
    def name(self):
        return self.object.name

    @abstractmethod
    def get_image_name(self):
        """Get image name"""

    @abstractmethod
    def get_tag(self):
        """Get image tag"""

    @abstractmethod
    def has_new_version(self):
        """Return true if there are a new version"""

    @abstractmethod
    def get_current_id(self):
        """Return the current id for notification"""

    @abstractmethod
    def get_latest_id(self):
        """Return the latest id for notification"""

    def _pull(self, name_with_tag):
        """Docker pull image tag"""
        self.logger.debug('Checking tag: %s', name_with_tag)
        try:
            if self.config.auth_json:
                self.client.login(self.config.auth_json.get("username"), self.config.auth_json.get("password"))

            # The authentification doesn't work with this call
            # See bugs https://github.com/docker/docker-py/issues/2225
            # return self.client.images.get_registry_data(tag)
            return self.client.images.pull(name_with_tag, auth_config=self.config.auth_json)
        except APIError as e:
            if '<html>' in str(e):
                self.logger.debug("Docker api issue. Ignoring")
                raise ConnectionError
            elif 'unauthorized' in str(e):
                raise ConnectionError
            elif 'Client.Timeout' in str(e):
                self.logger.critical(
                    "Couldn't find an image on docker.com for %s. Local Build?", name_with_tag)
                raise ConnectionError
            elif ('pull access' or 'TLS handshake') in str(e):
                self.logger.critical("Couldn't pull. Skipping. Error: %s", e)
                raise ConnectionError



from copy import deepcopy
from logging import getLogger
from os import environ
from pathlib import Path

from .logger import BlacklistFilter
from ..helpers.helpers import convert_to_boolean

MINIMUM_INTERVAL = 30

ENABLE_LABEL = "dockupdater.enable"
DISABLE_LABEL = "dockupdater.disable"

LABELS_MAPPING = {
    "dockupdater.latest": "latest",
    "dockupdater.notifiers": "notifiers",
    "dockupdater.stop_signal": "stop_signal",
    "dockupdater.cleanup": "cleanup",
    "dockupdater.template_file": "template_file",
    "dockupdater.wait": "wait",
    "dockupdater.recreate_first": "recreate_first"
}


class DefaultConfig(object):
    """Default configuration"""
    hostname = environ.get('HOSTNAME')
    interval = 300
    cron = None
    docker_sockets = ['unix://var/run/docker.sock']
    docker_tls = False
    docker_tls_verify = True
    log_level = 'info'
    cleanup = False
    run_once = False
    label = False
    stop_signal = None
    disable_containers_check = False
    disable_services_check = False
    latest = False
    wait = 0
    recreate_first = False

    repo_user = None
    repo_pass = None

    notifiers = []
    skip_start_notif = False
    template_file = None


class Config(object):
    def __init__(self, **kwargs):
        super().__setattr__("options", kwargs)
        self.logger = getLogger()
        self.compute_args()
        self.filtered_strings = None

    @classmethod
    def from_labels(cls, config, labels):
        """Create a new config object from an existing config and a dict of docker labels"""
        options = deepcopy(config.options)

        if labels:
            for label, value in labels.items():
                if label in LABELS_MAPPING:
                    if label in ["dockupdater.notifiers"]:
                        options[LABELS_MAPPING[label]] = value.split(" ")
                    elif label in ["dockupdater.latest", "dockupdater.cleanup"]:
                        options[LABELS_MAPPING[label]] = convert_to_boolean(value)
                    elif label in ["dockupdater.wait"]:
                        options[LABELS_MAPPING[label]] = int(value)
                    else:
                        options[LABELS_MAPPING[label]] = value
                    if label == "dockupdater.template_file":
                        # Reload template
                        options["template"] = Config.load_template(options.get('template_file'))

        return cls(**options)

    def __setattr__(self, key, value):
        if key in self.options:
            self.options[key] = value
        else:
            super().__setattr__(key, value)

    def __getattr__(self, attr):
        try:
            return self.options[attr]
        except KeyError:
            raise AttributeError

    def config_blacklist(self):
        """Mask sensitive data from logs"""
        filtered_strings = [getattr(self, key.lower()) for key, value in self.options.items()
                            if key.lower() in BlacklistFilter.blacklisted_keys]
        # Clear None values
        self.filtered_strings = list(filter(None, filtered_strings))
        # take lists inside of list and append to list
        for index, value in enumerate(self.filtered_strings, 0):
            if isinstance(value, list) or isinstance(value, tuple):
                self.filtered_strings.extend(self.filtered_strings.pop(index))
                self.filtered_strings.insert(index, self.filtered_strings[-1:][0])
        # Filter out no string item
        self.filtered_strings = [item for item in self.filtered_strings if isinstance(item, str)]
        # Added matching for ports
        ports = [string.split(':')[0] for string in self.filtered_strings if ':' in string]
        self.filtered_strings.extend(ports)
        # Added matching for tcp sockets. ConnectionPool ignores the tcp://
        tcp_sockets = [string.split('//')[1] for string in self.filtered_strings if '//' in string]
        self.filtered_strings.extend(tcp_sockets)
        # Get JUST hostname from tcp//unix
        for socket in getattr(self, 'docker_sockets'):
            self.filtered_strings.append(socket.split('//')[1].split(':')[0])

        for handler in self.logger.handlers:
            handler.addFilter(BlacklistFilter(set(self.filtered_strings)))

    def compute_args(self):
        if self.repo_user and self.repo_pass:
            self.options['auth_json'] = {'Username': self.repo_user, 'Password': self.repo_pass}
        else:
            self.options['auth_json'] = None

        if self.disable_containers_check and self.disable_services_check:
            raise AttributeError("Error you can't disable all monitoring (containers/services).")

        # Config sanity checks
        if self.cron:
            if not isinstance(self.cron, list):
                cron_times = self.cron.strip().split(' ')
                if len(cron_times) != 5:
                    self.logger.critical("Cron must be in cron syntax. e.g. * * * * * (5 places).")
                    raise AttributeError("Invalid cron")
                else:
                    self.logger.info("Cron configuration is valid. Using Cron schedule %s", cron_times)
                    self.cron = cron_times
                    self.interval = None
        else:
            if self.interval < MINIMUM_INTERVAL:
                self.logger.warning('Minimum value for interval was 30 seconds.')
                self.interval = MINIMUM_INTERVAL

        self.options['template'] = Config.load_template(self.template_file)

    @staticmethod
    def load_template(template_file):
        # Load default template file
        if not template_file:
            dir_path = Path().absolute()
            template_file = dir_path.joinpath("dockupdater/templates/notification.j2")

        if Path(template_file).exists():
            with open(template_file) as f:
                return f.read()
        else:
            raise AttributeError(f"Template file {template_file} not found")



from logging import getLogger
from os.path import isdir, isfile, join

from docker import DockerClient, tls


class Docker(object):
    """Class use to connect to the docker socket"""

    def __init__(self, socket, config, notification_manager):
        self.config = config
        self.socket = socket
        self.client = self.connect()
        self.logger = getLogger()

        self.notification_manager = notification_manager

    def connect(self):
        if self.config.docker_tls:
            try:
                cert_paths = {
                    'cert_top_dir': '/etc/docker/certs.d/',
                    'clean_socket': self.socket.split('//')[1]
                }
                cert_paths['cert_dir'] = join(cert_paths['cert_top_dir'], cert_paths['clean_socket'])
                cert_paths['cert_files'] = {
                    'client_cert': join(cert_paths['cert_dir'], 'client.cert'),
                    'client_key': join(cert_paths['cert_dir'], 'client.key'),
                    'ca_crt': join(cert_paths['cert_dir'], 'ca.crt')
                }

                if not isdir(cert_paths['cert_dir']):
                    self.logger.error('%s is not a valid cert folder', cert_paths['cert_dir'])
                    raise ValueError

                for cert_file in cert_paths['cert_files'].values():
                    if not isfile(cert_file):
                        self.logger.error('%s does not exist', cert_file)
                        raise ValueError

                tls_config = tls.TLSConfig(
                    ca_cert=cert_paths['cert_files']['ca_crt'],
                    verify=cert_paths['cert_files']['ca_crt'] if self.config.docker_tls_verify else False,
                    client_cert=(cert_paths['cert_files']['client_cert'], cert_paths['cert_files']['client_key'])
                )
                client = DockerClient(base_url=self.socket, tls=tls_config)
            except ValueError:
                self.logger.error('Invalid Docker TLS config for %s, reverting to unsecured', self.socket)
                client = DockerClient(base_url=self.socket)
        else:
            client = DockerClient(base_url=self.socket)

        return client



def set_properties(old, new, self_update=False):
    """Store object for spawning new container in place of the one with outdated image"""
    labels = old.attrs['Config']['Labels']
    ports = None if not old.attrs['Config'].get('ExposedPorts') else [
        (p.split('/')[0], p.split('/')[1]) for p in old.attrs['Config']['ExposedPorts'].keys()
    ]

    if self_update:
        if ports:
            labels["dockupdater.updater_port"] = ":".join(["{},{}".format(a, b) for a, b in ports])
            ports = None
        elif labels.get("dockupdater.updater_port"):
            ports = [(int(port.split(",")[0]), port.split(",")[1])
                     for port in labels.get("dockupdater.updater_port").split(":")]
            labels["dockupdater.updater_port"] = None
            del labels["dockupdater.updater_port"]

    properties = {
        'name': old.name,
        'hostname': old.attrs['Config']['Hostname'],
        'user': old.attrs['Config']['User'],
        'detach': True,
        'domainname': old.attrs['Config']['Domainname'],
        'tty': old.attrs['Config']['Tty'],
        'ports': ports,
        'volumes': None if not old.attrs['Config'].get('Volumes') else [
            v for v in old.attrs['Config']['Volumes'].keys()
        ],
        'working_dir': old.attrs['Config']['WorkingDir'],
        'image': new.tags[0],
        'command': old.attrs['Config']['Cmd'],
        'host_config': old.attrs['HostConfig'],
        'labels': labels,
        'entrypoint': old.attrs['Config']['Entrypoint'],
        'environment': old.attrs['Config']['Env']
    }

    return properties


def remove_sha_prefix(digest):
    if digest.startswith("sha256:"):
        return digest[7:]
    return digest


def convert_to_boolean(value):
    return str(value).lower() in ["yes", "y", "true"]


def get_id_from_image(image):
    return remove_sha_prefix(image.attrs.get('Image', image.attrs.get(id, image.id)))



from datetime import datetime, timezone, timedelta
from time import sleep

import click
from apscheduler.schedulers.background import BackgroundScheduler
from requests.exceptions import ConnectionError

from . import VERSION, BRANCH
from .lib.config import DefaultConfig, Config
from .lib.dockerclient import Docker
from .lib.logger import DockupdaterLogger
from .lib.notifiers import NotificationManager, StartupMessage
from .lib.scanner import Scanner


@click.command()
@click.version_option(version=VERSION)
@click.option("-d", "--docker-sockets", "docker_sockets", envvar="DOCKER_SOCKETS",
              default=DefaultConfig.docker_sockets, show_default=True, multiple=True,
              help='Socket for docker management. EXAMPLE: -d unix://var/run/docker.sock')
@click.option("-t", "--docker-tls", "docker_tls", envvar="DOCKER_TLS",
              is_flag=True, default=DefaultConfig.docker_tls, show_default=True,
              help='Enable docker TLS. REQUIRES: docker cert mount')
@click.option("-T", "--docker-tls-verify", "docker_tls_verify", envvar="DOCKER_TLS_VERIFY",
              is_flag=True, default=DefaultConfig.docker_tls_verify, show_default=True,
              help='Verify the CA Certificate mounted for TLS')
@click.option("-i", "--interval", "interval", envvar="INTERVAL",
              type=int, default=DefaultConfig.interval, show_default=True,
              help='Interval in seconds between checking for updates')
@click.option("-C", "--cron", "cron", envvar="CRON",
              default=DefaultConfig.cron,
              help='Cron formatted string for scheduling. EXAMPLE: "*/5 * * * *"')
@click.option("-l", "--log-level", "log_level", envvar="LOG_LEVEL",
              type=click.Choice(['debug', 'info', 'warn', 'error', 'critical']), default=DefaultConfig.log_level,
              show_default=True, help='Set logging level')
@click.option("-L", "--latest", "latest", envvar="LATEST",
              default=DefaultConfig.latest, show_default=True, is_flag=True,
              help='Pull the :latest tags and update all containers to it.')
@click.option("-o", "--run-once", "run_once", envvar="RUN_ONCE",
              is_flag=True, default=DefaultConfig.run_once, show_default=True,
              help='Single run')
@click.option("-N", "--notifiers", "notifiers", envvar="NOTIFIERS",
              default=DefaultConfig.notifiers, multiple=True,
              help='Apprise formatted notifiers\n'
                   'EXAMPLE: -N discord://1234123412341234/jasdfasdfasdfasddfasdf '
                   'mailto://user:pass@gmail.com')
@click.option("--skip-start-notif", "skip_start_notif", envvar="SKIP_START_NOTIF",
              default=DefaultConfig.skip_start_notif, is_flag=True, show_default=True,
              help='Skip notification of dockupdater has started')
@click.option("--template-file", "template_file", envvar="TEMPLATE_FILE",
              default=DefaultConfig.template_file, type=click.Path(),
              help='Use a custom template for notification')
@click.option("-k", "--label", "label", envvar="LABEL",
              default=DefaultConfig.label, is_flag=True, show_default=True,
              help='Enable label monitoring for dockupdater instead of monitoring all containers/services')
@click.option("-c", "--cleanup", "cleanup", envvar="CLEANUP",
              default=DefaultConfig.cleanup, is_flag=True, show_default=True,
              help='Remove old images after updating')
@click.option("-r", "--repo-user", "repo_user", envvar="REPO_USER",
              default=DefaultConfig.repo_user,
              help='Private docker registry username\n'
                   'EXAMPLE: foo@bar.baz')
@click.option("-R", "--repo-pass", "repo_pass", envvar="REPO_PASS",
              default=DefaultConfig.repo_pass,
              help='Private docker registry password')
@click.option("--stop-signal", "stop_signal", envvar="STOP_SIGNAL",
              type=int, default=DefaultConfig.stop_signal,
              help='Override the stop signal send to container')
@click.option("--disable-services-check", "disable_services_check", envvar="DISABLE_SERVICES_CHECK",
              default=DefaultConfig.disable_services_check, is_flag=True, show_default=True,
              help='Disable services (swarm) check')
@click.option("--disable-containers-check", "disable_containers_check", envvar="DISABLE_CONTAINERS_CHECK",
              default=DefaultConfig.disable_containers_check, is_flag=True, show_default=True,
              help='Disable standalone containers check')
@click.option("--hostname", "hostname",
              default=DefaultConfig.hostname, envvar="HOSTNAME",
              help='Set hostname (for debugging only)')
@click.option("-w", "--wait", "wait", envvar="WAIT",
              default=DefaultConfig.wait, type=click.INT,
              help='Define a time in seconds to wait after an update before updating any others containers/services.')
@click.option("-F", "--recreate-first", "recreate_first", envvar="RECREATE_FIRST",
              default=DefaultConfig.recreate_first, type=click.INT,
              help='Create a new container before stopping the old container.')
def cli(docker_sockets, docker_tls, docker_tls_verify, interval, cron, log_level, run_once, notifiers,
        skip_start_notif, label, cleanup, repo_user, repo_pass, stop_signal, disable_services_check,
        disable_containers_check, template_file, hostname, latest, wait, recreate_first):
    """Declare command line options"""

    # Create App logger
    log = DockupdaterLogger(level=log_level)
    log.logger.info('Version: %s-%s', VERSION, BRANCH)

    config = Config(docker_sockets=docker_sockets,
                    docker_tls=docker_tls,
                    docker_tls_verify=docker_tls_verify,
                    interval=interval,
                    cron=cron,
                    log_level=log_level,
                    run_once=run_once,
                    notifiers=notifiers,
                    skip_start_notif=skip_start_notif,
                    label=label,
                    cleanup=cleanup,
                    repo_user=repo_user,
                    repo_pass=repo_pass,
                    stop_signal=stop_signal,
                    disable_services_check=disable_services_check,
                    disable_containers_check=disable_containers_check,
                    template_file=template_file,
                    hostname=hostname,
                    latest=latest,
                    wait=wait,
                    recreate_first=recreate_first)
    config.config_blacklist()  # Configure mask on logger

    log.logger.debug("dockupdater configuration: %s", config.options)

    notification_manager = NotificationManager(config)
    scheduler = BackgroundScheduler()
    scheduler.start()

    for socket in config.docker_sockets:
        try:
            docker = Docker(socket, config, notification_manager)
            scanner = Scanner(docker)

            # Always begin to check the self update
            scanner.self_update()
            # Check docker swarm mode is running on a manager
            scanner.check_swarm_mode()

            if config.cron:
                scheduler.add_job(
                    scanner.update,
                    name=f'Cron container update for {socket}',
                    trigger='cron',
                    minute=config.cron[0],
                    hour=config.cron[1],
                    day=config.cron[2],
                    month=config.cron[3],
                    day_of_week=config.cron[4],
                    misfire_grace_time=20
                )
            else:
                if config.run_once:
                    scheduler.add_job(scanner.update, name=f'Run Once container update for {socket}')
                else:
                    scheduler.add_job(
                        scanner.update,
                        name=f'Initial run interval container update for {socket}'
                    )
                    scheduler.add_job(
                        scanner.update,
                        name=f'Interval container update for {socket}',
                        trigger='interval', seconds=config.interval,
                        misfire_grace_time=20
                    )
        except ConnectionError:
            log.logger.error("Could not connect to socket %s. Check your config", config.socket)

    if config.run_once:
        next_run = None
    elif config.cron:
        next_run = scheduler.get_jobs()[0].next_run_time
    else:
        now = datetime.now(timezone.utc).astimezone()
        next_run = (now + timedelta(0, config.interval)).strftime("%Y-%m-%d %H:%M:%S")

    if not config.skip_start_notif:
        notification_manager.send(StartupMessage(config.hostname, next_run=next_run), config.notifiers)

    while scheduler.get_jobs():
        sleep(10)

    scheduler.shutdown()


def main():
    cli()


if __name__ == "__main__":
    main()



from .abstract import AbstractObject
from ..helpers.helpers import remove_sha_prefix
from ..lib.config import Config


class Service(AbstractObject):
    def __init__(self, docker_client, service):
        super().__init__(docker_client, service)
        self._latest_sha = None
        self._current_sha = None

    def get_image_name(self):
        return self.service.attrs['Spec']['TaskTemplate']['ContainerSpec']['Image'].split(":", 1)[0]

    def get_tag(self):
        tag = self.service.attrs['Spec']['TaskTemplate']['ContainerSpec']['Image'].split(":", 1)
        if len(tag) == 2:
            tag = tag[1]
            if ":" in tag and "@" in tag:
                tag = tag.split("@")[0]
            return tag
        return "latest"

    def get_sha(self):
        tag = self.service.attrs['Spec']['TaskTemplate']['ContainerSpec']['Image'].split(":", 1)
        if len(tag) == 2:
            tag = tag[1]
            sha = ""
            if ":" in tag and "@" in tag:
                sha = tag.split("@")[1]
            return remove_sha_prefix(sha)
        return ""  # Empty, force to update this service for next time

    @property
    def service(self):
        return self.object

    def get_current_id(self):
        return (self._current_sha or "")[:10]

    def get_latest_id(self):
        return (self._latest_sha or "")[:10]

    def has_new_version(self):
        self.config = Config.from_labels(self.config, self.service.attrs.get('Spec', dict()).get('Labels'))

        current_image_name = self.get_image_name()
        current_tag = self.get_tag()

        self.logger.debug("%s:%s", current_image_name, current_tag)

        current_sha = self.get_sha()
        if not current_sha:
            self.logger.error('No image SHA for %s. Update it for next time', current_image_name)

        if self.config.latest:
            current_tag = "latest"

        try:
            latest_image = self._pull(f"{current_image_name}:{current_tag}")
        except ConnectionError:
            return False

        latest_sha = self._get_digest(latest_image)
        self._latest_sha = latest_sha
        self._current_sha = current_sha
        return current_sha != latest_sha

    def update(self):
        if not self._latest_sha:
            self.has_new_version()

        self.logger.info('%s will be updated', self.name)
        self.service.update(image=f"{self.get_image_name()}:{self.get_tag()}@sha256:{self._latest_sha}")

    def _get_digest(self, image):
        digest = image.attrs.get(
            "Descriptor", {}
        ).get("digest") or image.attrs.get(
            "RepoDigests"
        )[0].split('@')[1] or image.id

        return remove_sha_prefix(digest)



from abc import ABC
from datetime import datetime, timezone
from logging import getLogger

import apprise
from jinja2 import Environment, BaseLoader

from ..update.service import Service


class BaseMessage(ABC):
    """Abstract message to send with apprise"""

    def __init__(self, title, body):
        self.title = title
        self.body = body


class StartupMessage(BaseMessage):
    def __init__(self, hostname, next_run=None):
        now = datetime.now(timezone.utc).astimezone()
        title = f'dockupdater has started'
        body_fields = [
            f'Host: {hostname}',
            f'Time: {now.strftime("%Y-%m-%d %H:%M:%S")}',
            f'Next Run: {next_run}']
        super().__init__(title, '\n'.join(body_fields))


class TemplateMessage(BaseMessage):
    """Load and parse message to send from template file"""

    def __init__(self, container_or_service):
        if isinstance(container_or_service, Service):
            title = f'dockupdater has updated services!'
        else:
            title = f'dockupdater has updated containers!'

        template = Environment(loader=BaseLoader).from_string(container_or_service.config.template)
        body = template.render(object=container_or_service)
        super().__init__(title, body)


class NotificationManager(object):
    def __init__(self, config):
        self.config = config
        self.logger = getLogger()

    def build_apprise(self, notifiers):
        asset = apprise.AppriseAsset()
        asset.app_id = "dockupdater"
        asset.app_desc = "dockupdater"
        asset.app_url = "https://github.com/dockupdater/dockupdater"
        asset.html_notify_map['info'] = '#5F87C6'

        apprise_obj = apprise.Apprise(asset=asset)

        for notifier in notifiers:
            if notifier:
                add = apprise_obj.add(notifier)
                if not add:
                    self.logger.error('Could not add notifier %s', notifier)

        return apprise_obj

    def send(self, message, notifiers):
        apprise = self.build_apprise(notifiers)

        if apprise.servers:
            apprise.notify(title=message.title, body=message.body)



from logging import Filter, getLogger, Formatter, StreamHandler


class BlacklistFilter(Filter):
    """
    Log filter for blacklisted tokens and passwords
    """

    blacklisted_keys = ['repo_user', 'repo_pass', 'auth_json', 'docker_sockets', 'notifiers']

    def __init__(self, filteredstrings):
        super().__init__()
        self.filtered_strings = filteredstrings

    def filter(self, record):
        for item in self.filtered_strings:
            try:
                if item in record.msg:
                    record.msg = record.msg.replace(item, 8 * '*' + item[-5:])
                if any(item in str(arg) for arg in record.args):
                    record.args = tuple(arg.replace(item, 8 * '*' + item[-5:]) if isinstance(arg, str) else arg
                                        for arg in record.args)
            except TypeError:
                pass
        return True


class DockupdaterLogger(object):
    def __init__(self, level='INFO'):
        # Create the Logger
        self.logger = getLogger()
        try:
            self.logger.setLevel(level.upper())
        except ValueError:
            level = "INFO"
            self.logger.setLevel(level.upper())

        # Create a Formatter for formatting the log messages
        logger_formatter = Formatter('%(asctime)s : %(levelname)s : %(module)s : %(message)s', '%Y-%m-%d %H:%M:%S')

        # Add the console logger
        console_logger = StreamHandler()
        console_logger.setFormatter(logger_formatter)

        console_logger.setLevel(level.upper())

        # Add the Handler to the Logger
        self.logger.addHandler(console_logger)

        # Less verbose apscheduler logging if info
        if self.logger.getEffectiveLevel() == 20:
            getLogger('apscheduler').setLevel('WARNING')



from time import sleep

from docker.errors import APIError, NotFound

from .abstract import AbstractObject
from ..helpers.helpers import set_properties, remove_sha_prefix, get_id_from_image
from ..lib.config import Config


class Container(AbstractObject):
    def __init__(self, docker_client, container):
        super().__init__(docker_client, container)
        self._latest_image = None
        self._current_id = None

    def get_image_name(self):
        return self.container.attrs['Config']['Image'].split(":", 1)[0]

    def get_tag(self):
        image = self.container.attrs['Config']['Image']
        if ":" in image:
            return image.split(":", 1)[1]
        else:
            tags = self.client.images.get(self.get_image_name()).tags
            if tags:
                tag = tags[0]
                if ":" in tag:
                    return tag.split(":", 1)[1]
                else:
                    return "latest"

    @property
    def container(self):
        return self.object

    def get_current_id(self):
        return (self._current_id or "")[10:]

    def get_latest_id(self):
        if self._latest_image:
            return get_id_from_image(self._latest_image)[10:]
        return ""

    def is_dockupdater(self):
        dockupdater = "dockupdater" in self.container.attrs.get("Config", dict()).get("Image", self.name)
        if not dockupdater:
            for history in self.container.image.history():
                if "dockupdater" in (history.get("Tags", list()) or list()):
                    dockupdater = True
                    break
        if dockupdater:
            return True
        return False

    def has_new_version(self):
        self.config = Config.from_labels(self.config, self.container.labels)

        current_image_name = self.get_image_name()
        current_tag = self.get_tag()
        self._current_id = remove_sha_prefix(self.container.attrs.get('Image', self.container.id))

        if self.config.latest:
            current_tag = "latest"

        try:
            latest_image = self._pull(f"{current_image_name}:{current_tag}")
            self._latest_image = latest_image
        except ConnectionError:
            return False

        latest_id = get_id_from_image(latest_image)

        return self._current_id != latest_id

    def update(self):
        if self.is_dockupdater():
            self.logger.info("Dockupdater container is ready to update")
            self.config.recreate_first = True  # Always recreate dockupdater container first
        elif self.container.attrs['Config'].get('ExposedPorts') and self.config.recreate_first:
            self.config.recreate_first = False
            self.logger.warning(
                "Option recreate_first isn't compatible when container has exposed ports. Option is set to False."
            )

        if not self._current_id or not self._latest_image:
            self.has_new_version()

        if self.config.recreate_first:
            new_name = f"{self.name}_old"
            if self.is_dockupdater():
                new_name = f"{self.name}_old_dockupdater"
            self.container.rename(new_name)
        else:
            self.stop()
            self.remove()

        self.logger.info('%s will be updated', self.container.name)
        self.recreate()

        if self.is_dockupdater():
            self.logger.info('Waiting for new dockupdater container')
            sleep(600)

        if self.config.recreate_first:
            self.stop()
            self.remove()

        if self.config.cleanup:
            try:
                self.client.images.remove(self._current_id)
            except APIError as e:
                self.logger.error("Could not delete old image for %s, Error: %s", self.container.name, e)

    def stop(self):
        self.logger.debug('Stopping container: %s', self.object.name)

        if self.config.stop_signal:
            try:
                self.container.kill(signal=self.config.stop_signal)
            except APIError as e:
                self.logger.error('Cannot kill container using signal %s. stopping normally. Error: %s',
                                  self.object.stop_signal, e)
                self.container.stop()
        else:
            self.container.stop()

    def remove(self):
        self.logger.debug('Removing container: %s', self.container.name)
        if not self.container.attrs.get('HostConfig', dict()).get('AutoRemove'):
            try:
                self.container.remove()
            except NotFound as e:
                self.logger.error("Could not remove container. Error: %s", e)
                return
        else:
            # Docker will remove container, but some time that take few seconds
            try:
                while True:
                    self.client.containers.get(self.name)
            except NotFound:
                pass

    def recreate(self):
        new_config = set_properties(
            old=self.container,
            new=self._latest_image,
            self_update=self.is_dockupdater()
        )
        created = self.client.api.create_container(**new_config)
        new_container = self.client.containers.get(created.get("Id"))

        # connect the new container to all networks of the old container
        for network_name, network_config in self.container.attrs['NetworkSettings']['Networks'].items():
            network = self.client.networks.get(network_config['NetworkID'])
            try:
                network.disconnect(new_container.id, force=True)
            except APIError:
                pass
            new_network_config = {
                'container': new_container,
                'aliases': network_config['Aliases'],
                'links': network_config['Links'],
                'ipv4_address': network_config['IPAddress'],
                'ipv6_address': network_config['GlobalIPv6Address']
            }
            try:
                network.connect(**new_network_config)
            except APIError as e:
                if any(err in str(e) for err in ['user configured subnets', 'user defined networks']):
                    if new_network_config.get('ipv4_address'):
                        del new_network_config['ipv4_address']
                    if new_network_config.get('ipv6_address'):
                        del new_network_config['ipv6_address']
                    network.connect(**new_network_config)
                else:
                    self.logger.error('Unable to attach updated container to network "%s". Error: %s', network.name, e)

        new_container.start()



from setuptools import setup, find_packages

from dockupdater import VERSION

requirements = ['docker>=3.7.0',
                'apscheduler>=3.5.3',
                'requests>=2.21.0',
                'apprise>=0.5.2',
                'jinja2>=2.10',
                'click>=7.0']


def read(filename):
    with open(filename) as f:
        return f.read()


setup(
    name='dockupdater',
    version=VERSION,
    maintainer='Mathieu Cantin',
    maintainer_email='harcher81@gmail.com',
    description='Automatically keep your docker services and your docker containers up-to-date with the latest version',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    url='https://dockupdater.dev',
    license='MIT',
    classifiers=['Programming Language :: Python',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7'],
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    include_package_data=True,
    scripts=['entrypoint'],
    install_requires=requirements,
    python_requires='>=3.6.2'
)
